from abc import ABC, abstractmethod

import numpy as np


class DynamicsModel(ABC):
    """Baseclass for dynamics models.

    Defines an `predict` method.
    """

    @abstractmethod
    def predict(self, x: np.ndarray, u: np.ndarray, t: int = None, noise: np.ndarray = None):
        """Predict the next state.

        For time-varying models `t` is required.

        Args:
            x: (dX,) Current state
            u: (dU,) Current action
            t: Current timestep, may be `None`
            noise: (dX,) State noise, may be `None` to sample without noise

        Returns:
            x: (dX,) Next state
        """



import json
from typing import List

from donk.datalogging.datalogging import DataLogger

TODO!!!!!
Converty types to python primites o the fly


class JSONDataLogger(DataLogger):
    """Log data as single JSON file."""

    def __init__(self, file) -> None:
        super().__init__()

        self.file = file
        self.data = {}

    def log(self, key: List[str], data):
        """Log one data.

        Args:
            key: Path-like, hierarchical name associated with the data
            data: Data to log, should be python primitives
        """
        # Find/create nested dict
        d = self.data
        for key_part in key[:-1]:
            if key_part not in d:
                d[key_part] = {}
            d = d[key_part]

        # Store data
        d[key[-1]] = data

    def flush(self):
        """Write data to the file."""
        with open(self.file, "w") as f:
            json.dump(self.data, f, separators=(',', ':'))



from __future__ import annotations

import numpy as np

from donk.costs.cost_function import CostFunction
from donk.samples.traj_dist import TrajectoryDistribution
from donk.utils import trace_of_product


class QuadraticCosts(CostFunction):
    """Quadratic cost function.

    cost(x, u) = 1/2 [x u]^T*C*[x u] + [x u]^T*c + cc
    """

    def __init__(self, C: np.ndarray, c: np.ndarray, cc: np.ndarray):
        """Initialize this LinearDynamics object.

        Args:
            C: (..., dX+dU, dX+dU), Quadratic term
            c: (..., dX+dU), Linear term
            cc: (...,), Constant term

        """
        T = c.shape[:-1]
        dXU = c.shape[-1]

        # Check shapes
        assert C.shape == (*T, dXU, dXU), f"{C.shape} != {(*T, dXU, dXU)}"
        assert c.shape == (*T, dXU), f"{c.shape} != {(*T, dXU)}"
        assert cc.shape == (*T, ), f"{cc.shape} != {(*T, )}"

        self.C = C
        self.c = c
        self.cc = cc

    def __add__(self, other) -> QuadraticCosts:
        """Sum two cost functions."""
        if isinstance(other, QuadraticCosts):
            return QuadraticCosts(self.C + other.C, self.c + other.c, self.cc + other.cc)
        return NotImplemented

    def __mul__(self, other) -> QuadraticCosts:
        """Scale cost function with  vonstant scalar."""
        if np.isscalar(other):
            return QuadraticCosts(self.C * other, self.c * other, self.cc * other)
        return NotImplemented

    def __rmul__(self, other) -> QuadraticCosts:
        """Scale cost function with  vonstant scalar."""
        return self.__mul__(other)

    def compute_costs(self, X: np.ndarray, U: np.ndarray) -> np.ndarray:
        """Evaluate costs for trajectories.

        Args:
            X: (..., T+1, dX), states
            U: (..., T, dU), actions

        Returns:
            costs: (..., T+1), Costs at each time step
        """
        dU = U.shape[-1]
        # Add a dummy zero action to create a rectangular array
        traj = np.concatenate([
            X,
            np.concatenate([
                U,
                np.zeros(U.shape[:-2] + (1, dU)),
            ], axis=-2),
        ], axis=-1)

        costs = (
            # Quadratic costs
            # 1/2 * mu^t C mu
            np.einsum("...i,...ij,...j->...", traj, self.C, traj) / 2 +
            # Linear costs
            # mu^T c
            np.sum(traj * self.c, axis=-1) +
            # Constant costs
            # cc
            self.cc
        )
        return costs

    def quadratic_approximation(self, X: np.ndarray, U: np.ndarray) -> QuadraticCosts:
        """Compute a quadratic approximation (2nd order Taylor) at the given trajectory.

        This function is already quadratic, thus identical to it's quadratic approximation.
        The location of the trarectory does not change the approximation, thus is ignored.

        Args:
            X: (T+1, dX), states. Ignored
            U: (T, dX), actions. Ignored
        """
        return self

    def expected_costs(self, traj: TrajectoryDistribution) -> np.ndarray:
        """Compute estimated costs for trajectory distribution under quadratic cost approximation.

        1/2 * (Tr(C Sigma) + mu^t C mu) + mu^T c + cc

        Args:
            traj_mean: (..., dXU), Mean of trajectory distribution
            traj_covar: (..., dXU), Covariances of trajectory distribution

        Returns:
            expectation: (..., ) Expected costs at each time step
        """
        expectation = self.compute_costs(traj.X_mean, traj.U_mean) + trace_of_product(self.C, traj.covar) / 2
        return expectation

    @staticmethod
    def quadratic_cost_approximation_l2(t: np.ndarray, w: np.ndarray) -> QuadraticCosts:
        """Constructs a quadratic cost function for a non-equilibrium L2 loss.

        L(xu) = 0.5*sum_i w_i*(t_i-xu_i)**2

        Actually, this is not an approximation. Rather it's exact.

        Args:
            t: (..., dXU), Target for each state/action at each timestep.
            w: (..., dXU), Weight for each state/action at each timestep.
        """
        dX = t.shape[-1]

        C = np.einsum("...i,ij->...ij", w, np.eye(dX))
        c = -np.einsum("...i,...i->...i", w, t)
        cc = np.einsum("...i,...i,...i->...", t, w, t) / 2

        return QuadraticCosts(C, c, cc)

    @staticmethod
    def quadratic_cost_approximation_l1(xu: np.ndarray, t: np.ndarray, w: np.ndarray, alpha: float) -> QuadraticCosts:
        """Constructs a quadratic cost function approximation for a non-equilibrium L1 loss.

        L(xu) = sum_i w_i*sqrt((t_i-xu_i)**2 + alpha)

        For sake of differentiability, a small constant alpha is added.

        Args:
            xu: (..., dX+dU), Current trajectory
            t: (..., dX+dU), Target for each state/action at each timestep.
            w: (..., dX+dU), Weight for each state/action at each timestep.
            alpha: Small constant to create smooth differentials
        """
        dXU = xu.shape[-1]

        abs_squared = (t - xu)**2 + alpha
        abs_d = np.sqrt(abs_squared)
        abs_d_cubed = abs_squared * abs_d

        # C = diag(alpha * w / )
        C = np.einsum("...i,ij->...ij", alpha * w / abs_d_cubed, np.eye(dXU))
        c = w * ((xu - t) / abs_d - alpha * xu / abs_d_cubed)
        cc = np.sum(w * (abs_d + xu * (t - xu) / abs_d + alpha * xu**2 / abs_d_cubed / 2), axis=-1)

        return QuadraticCosts(C, c, cc)



from donk.policy.lin_gaussian import LinearGaussianPolicy
from donk.policy.noise import smooth_noise
from donk.policy.policy import Policy

__all__ = [
    "Policy",
    "LinearGaussianPolicy",
    "smooth_noise",
]



import numpy as np

import donk.datalogging as datalogging
from donk.costs import CostFunction, QuadraticCosts
from donk.dynamics import LinearDynamics, linear_dynamics
from donk.dynamics.prior import DynamicsPrior
from donk.policy import LinearGaussianPolicy
from donk.samples import StateDistribution
from donk.traj_opt import ILQR, lqg


class TrajOptAlgorithm:
    """Algorithm for iterative LQR trajectory optimization."""

    def __init__(
        self, kl_step: float, max_step_mult: float = 10, min_step_mult: float = 0.1, dynamics_regularization: float = 1e-6
    ) -> None:
        self.kl_step = kl_step
        self.kl_step_mult: float = 1
        self.max_step_mult: float = max_step_mult
        self.min_step_mult: float = min_step_mult
        self.dynamics_regularization: float = dynamics_regularization

        self.pol: LinearGaussianPolicy = None
        self.dyn: LinearDynamics = None
        self.x0: StateDistribution = None
        self.costs: QuadraticCosts = None
        self.ilqr: ILQR = None

    def iteration(
        self,
        X: np.ndarray,
        U: np.ndarray,
        cost_function: CostFunction,
        prev_pol: LinearGaussianPolicy = None,
        prior: DynamicsPrior = None,
    ):
        """Perform one iteration of trajectory optimization.

        Args:
            X: (N, T+1, dX) Train states
            U: (N, T, dU) Train actions
            cost_function: Cost function to minimize
            prev_pol: Policy to constrain trajectory optimization to
            prior: Optional prior for dynamics fit
        """
        prev_dyn = self.dyn
        prev_x0 = self.x0
        prev_costs = self.costs
        prev_pol = prev_pol if prev_pol is not None else self.pol

        # Fit initial state distribution
        self.x0 = StateDistribution.fit(X[:, 0])

        # Fit dynamics
        self.dyn = linear_dynamics.fit_lr(X, U, prior=prior, regularization=self.dynamics_regularization)

        # Evaluate costs
        # Assume trajectory mean as operating point
        self.costs = cost_function.quadratic_approximation(np.mean(X, axis=0), np.mean(U, axis=0))

        # Perform trajectory optimization
        ilqr = ILQR(self.dyn, prev_pol, self.costs, self.x0)
        self.pol = ilqr.optimize(self.kl_step * self.kl_step_mult).policy

        # Log
        datalogging.log(
            prev_pol=prev_pol,
            pol=self.pol,
            dyn=self.dyn,
            x0=self.x0,
            costs=self.costs,
            kl_step=self.kl_step * self.kl_step_mult,
        )

        # Update kl_step
        if prev_dyn is not None:
            with datalogging.Context("step_adjust"):
                self.kl_step_mult = lqg.step_adjust(
                    self.kl_step_mult, self.dyn, self.pol, self.x0, self.costs, prev_dyn, prev_pol, prev_x0, prev_costs, self.max_step_mult,
                    self.min_step_mult
                )



import numpy as np

from donk.policy.lin_gaussian import LinearGaussianPolicy


def constant_policy(T: int, dX: int, u: np.ndarray, variance: np.ndarray) -> LinearGaussianPolicy:
    """Generate an initial policy of constant actions with added noise.

    Args:
        T: Number of time steps
        dX: Size of state space
        u: (dU, ) Constant action to perform every step
        variance: (dU, ) Diagonal action variance
    """
    u, variance = np.asarray(u), np.asarray(variance)
    dU = u.shape[0]

    return LinearGaussianPolicy(
        K=np.zeros((T, dU, dX)),
        k=np.tile(u, (T, 1)),
        covar=np.tile(np.diag(variance), (T, 1, 1)),
        inv_covar=np.tile(np.diag(1 / variance), (T, 1, 1)),
    )



from dataclasses import dataclass

import numpy as np
from scipy import optimize
from scipy.linalg import solve

import donk.datalogging as datalogging
from donk.costs.quadratic_costs import QuadraticCosts
from donk.dynamics import LinearDynamics
from donk.policy import LinearGaussianPolicy
from donk.samples import StateDistribution
from donk.samples.traj_dist import TrajectoryDistribution
from donk.utils import regularize, symmetrize, trace_of_product


@dataclass
class ILQRStepResult:
    eta: float
    policy: LinearGaussianPolicy
    kl_div: float
    expected_costs: float
    trajectory: TrajectoryDistribution


class ILQR:
    """iLQG trajectory optimization."""

    def __init__(
        self,
        dynamics: LinearDynamics,
        prev_pol: LinearGaussianPolicy,
        costs: QuadraticCosts,
        initial_state: StateDistribution,
    ) -> None:
        """Initialize this object.

        Args:
            dynamics: Time-varying linear Gaussian dynamics model
            prev_pol: Policy to contrain the new policy to
            initial_state: initial state distribution mean
            costs: (T, dX+dU, dX+dU): Quadratic cost function
        """
        self.dynamics = dynamics
        self.prev_pol = prev_pol
        self.costs = costs
        self.initial_state = initial_state

        # Compute extended costs
        self.C_kl, self.c_kl = extended_costs_kl(prev_pol)

    def step(self, eta: float) -> ILQRStepResult:
        """Perfrom the unconstrained optimization under the given Lagrange multiplier.

        Args:
            eta: Lagrange multiplier
        """
        # Compute extended costs for the given Lagrangian multiplier
        C_ext = self.costs.C / eta
        C_ext[:-1] += self.C_kl
        c_ext = self.costs.c / eta
        c_ext[:-1] += self.c_kl

        # Perform unconstrained trajectory optimization under the extended costs
        pol = backward(self.dynamics, C_ext, c_ext)

        # Compute KL-divergence and expected costs of the new policy
        traj = forward(self.dynamics, pol, self.initial_state)
        kl_div = kl_divergence_action(traj.X_mean, pol, self.prev_pol)
        expected_costs = np.sum(self.costs.expected_costs(traj))

        return ILQRStepResult(eta, pol, kl_div, expected_costs, traj)

    def sample_surface(self, min_eta: float = 1e-6, max_eta: float = 1e16, N: int = 100):
        """Sample the Lagrangian at different values for eta.

        For visualization/debugging purposes.

        Args:
            min_eta: Minimal value of the Lagrangian multiplier
            max_eta: Maximal value of the Lagrangian multiplier
            N: Number of samples. Samples are equally distanced in log space to base 10.
        """
        results = [self.step(eta) for eta in np.logspace(np.log10(min_eta), np.log10(max_eta), num=N)]
        return results

    def optimize(self, kl_step: float, min_eta: float = 1e-6, max_eta: float = 1e16, rtol: float = 1e-2):
        """Perform iLQG trajectory optimization.

        Args:
            kl_step: KL divergence threshold to previous policy
            min_eta: Minimal value of the Lagrangian multiplier
            max_eta: Maximal value of the Lagrangian multiplier
            rtol: Tolerance of found solution to kl_step. Levine et al. propose a value of 0.1 in "Learning Neural Network Policies with
                  Guided Policy Search under Unknown Dynamics", chapter 3.1
            full_history: Whether to return ahistory of all optimization steps, for debug purposes

        Returns:
            result: A `ILQRStepResult` or
                    a list of `ILQRStepResult` if `full_history` is enabled (in order they were visited)
        """
        # Check if constraind is fulfilled at maximum deviation
        if self.step(min_eta).kl_div <= kl_step:
            return self.step(min_eta)

        # Check if constraint cen be fulfilled at all
        if self.step(max_eta).kl_div > kl_step:
            raise ValueError(f"max_eta eta to low ({max_eta})")

        # Find the point where kl divergence equals the kl_step
        def constraint_violation(log_eta):
            return self.step(np.exp(log_eta)).kl_div - kl_step

        # Search root of the constraint violation
        # Perform search in log-space, as this requires much fewer iterations
        log_eta = optimize.brentq(constraint_violation, np.log(min_eta), np.log(max_eta), rtol=rtol)

        return self.step(np.exp(log_eta))


def backward(dynamics: LinearDynamics, C, c, gamma=1) -> LinearGaussianPolicy:
    """Perform LQR backward pass.

    `C` is required to be symmetric.
    `C[dU:, dU:]` is required to be s.p.d.
    At `C[T]` only `C[T, :dX, :dX]` needs to be defined. The other costs at time `T+1` can be undefined.

    Args:
        dynamics: A LinearDynamics object.
        C: (T+1, dX+dU, dX+dU) Quadratic costs
        c: (T+1, dX+dU) Linear costs
        gamma: Discount factor for future rewards

    Returns:
        traj_distr: A new linear Gaussian policy.

    """
    T, dX, dU = dynamics.T, dynamics.dX, dynamics.dU

    K = np.empty((T, dU, dX))
    k = np.empty((T, dU))
    pol_covar = np.empty((T, dU, dU))
    inv_pol_covar = np.empty((T, dU, dU))

    # Set value of final state
    V = C[T, :dX, :dX]
    v = c[T, :dX]

    # For convenicence
    F, f = dynamics.F, dynamics.f

    # Compute state-action-state function at each time step.
    for t in reversed(range(T)):
        # Compute Q function.
        Q = C[t] + gamma * F[t].T @ V @ F[t]
        q = c[t] + gamma * F[t].T @ (V @ f[t] + v)
        symmetrize(Q)

        # Compute inverse of Q function action component (as it's required explicitely anyway).
        Q_uu_inv = solve(Q[dX:, dX:], np.eye(dU), assume_a="pos")
        symmetrize(Q_uu_inv)

        # Compute controller parameters
        # K_t = -Q_{uut}^-1 * Q_{uxt}
        K[t] = -Q_uu_inv @ Q[dX:, :dX]
        # K_t = -Q_{uut}^-1 * q_{ut}
        k[t] = -Q_uu_inv @ q[dX:]
        # pol_covar_t = Q_{uut}^-1
        pol_covar[t] = Q_uu_inv
        inv_pol_covar[t] = Q[dX:, dX:]

        # Compute value function.
        V = Q[:dX, :dX] + Q[:dX, dX:] @ K[t]
        v = q[:dX] + Q[:dX, dX:] @ k[t]
        symmetrize(V)

    return LinearGaussianPolicy(K, k, pol_covar, inv_pol_covar)


def forward(
    dynamics: LinearDynamics,
    policy: LinearGaussianPolicy,
    initial_state: StateDistribution,
    regularization=1e-6,
) -> TrajectoryDistribution:
    """Perform LQR forward pass.

    Computes state-action marginals from dynamics and policy.

    Distribution of last time step only contains the state distribution, as there is no corresponding action.

    Args:
        dynamics: A LinearGaussianPolicy.
        policy: A LinearDynamics.
        X_0_mean: (dX, ) Mean of initial state distribution.
        X_0_covar: (dX, dX) Covariance of initial state distribution.

    Returns:
        traj_mean: (T+1, dX+dU) mean state-action vectors.
        traj_covar: (T+1, dX+dU, dX+dU) state-action covariance matrices.

    """
    T, dX, dU = dynamics.T, dynamics.dX, dynamics.dU

    traj_mean = np.empty((T + 1, dX + dU))
    traj_covar = np.empty((T + 1, dX + dU, dX + dU))

    # Set initial state dist
    traj_mean[0, :dX] = initial_state.mean
    traj_covar[0, :dX, :dX] = initial_state.covar

    # Set action part for final state
    traj_mean[T, dX:] = 0
    traj_covar[T] = 0

    # For convenicence
    F, f, dyn_covar = dynamics.F, dynamics.f, dynamics.covar
    K, k, pol_covar = policy.K, policy.k, policy.covar

    for t in range(T):
        # u_t = K_t*x_t + k_t
        traj_mean[t, dX:] = K[t] @ traj_mean[t, :dX] + k[t]
        # TODO Formulas
        traj_covar[t, dX:, :dX] = K[t] @ traj_covar[t, :dX, :dX]
        traj_covar[t, :dX, dX:] = traj_covar[t, dX:, :dX].T
        traj_covar[t, dX:, dX:] = traj_covar[t, dX:, :dX] @ K[t].T + pol_covar[t]

        # x_t+1 = Fm_t*[x_t;u_t] + fv_t
        traj_mean[t + 1, :dX] = F[t] @ traj_mean[t] + f[t]
        # TODO Formula
        traj_covar[t + 1, :dX, :dX] = F[t] @ traj_covar[t] @ F[t].T + dyn_covar[t]

        symmetrize(traj_covar[t])
        regularize(traj_covar[t], regularization)

    # Symmetrize state distribution of final state
    symmetrize(traj_covar[T, :dX, :dX])

    return TrajectoryDistribution(traj_mean, traj_covar, dX, dU)


def extended_costs_kl(prev_pol: LinearGaussianPolicy):
    """Compute expansion of extended cost used in the iLQR backward pass.

    The extended cost function is -log p(u_t | x_t) with p being the previous trajectory distribution.
    Thus, rewarding similarity of actions to the previous policy.

    Returns:
        C: Quadratic term of extended costs
        c: Linear term of extended costs
    """
    T, dX, dU = prev_pol.T, prev_pol.dX, prev_pol.dU

    C = np.empty((T, dX + dU, dX + dU))
    c = np.empty((T, dX + dU))

    # For convenicence
    K, k, inv_pol_covar = prev_pol.K, prev_pol.k, prev_pol.inv_covar

    for t in range(T):
        C[t, :dX, :dX] = K[t].T @ inv_pol_covar[t] @ K[t]
        C[t, :dX, dX:] = -K[t].T @ inv_pol_covar[t]
        C[t, dX:, :dX] = -inv_pol_covar[t] @ K[t]
        C[t, dX:, dX:] = inv_pol_covar[t]
        c[t, :dX] = K[t].T @ inv_pol_covar[t] @ k[t]
        c[t, dX:] = -inv_pol_covar[t] @ k[t]

    return C, c


def kl_divergence_action(X, pol: LinearGaussianPolicy, prev_pol: LinearGaussianPolicy):
    """Compute KL divergence between new and previous trajectory distributions.

    Args:
        X: (T, dX), mean of states to compare actions for.
        pol: LinearGaussianPolicy, new policy.
        prev_pol: LinearGaussianPolicy, previous policy.

    Returns:
        kl_div: The mean KL divergence between the new and previous actions over time.

    See:
        https://web.stanford.edu/~jduchi/projects/general_notes.pdf Chapter 9
    """
    T, dU = pol.T, pol.dU

    kl_div = 0
    for t in range(T):
        delta_u = (prev_pol.K[t] - pol.K[t]) @ X[t] + prev_pol.k[t] - pol.k[t]
        kl_div += 0.5 * (
            # tr(sigma_1^-1 * sigma_0)
            trace_of_product(prev_pol.inv_covar[t], pol.covar[t]) +
            # (mu_1 - mu_0)^T * sigma_1^-1 * (mu_1 - mu_0)
            delta_u.T @ prev_pol.inv_covar[t] @ delta_u +
            # -k
            -dU +
            # log(det(sigma_1) / det(sigma_0))
            2 * sum(np.log(np.diag(prev_pol.chol_covar[t])) - np.log(np.diag(pol.chol_covar[t])))
        )

    return kl_div / T


def step_adjust(
    step_mult: float,
    dyn: LinearDynamics,
    policy: LinearGaussianPolicy,
    x0: StateDistribution,
    costs: QuadraticCosts,
    dyn_prev: LinearDynamics,
    policy_prev: LinearGaussianPolicy,
    x0_prev: StateDistribution,
    costs_prev: QuadraticCosts,
    max_step_mult: float = 10,
    min_step_mult: float = 0.1,
):
    """Compute new multiplier for KL divergence constraint based on expected vs. actual improvement.

    See:
        Montgomery et al. "Guided Policy Search as Approximate Mirror Descent", Sec. 4.2
    """
    # Cost under previous dynamics and previous policy
    previous_costs = np.sum(costs_prev.expected_costs(forward(dyn_prev, policy_prev, x0_prev)))

    # Cost under previous dynamics and new policy
    predicted_new_costs = np.sum(costs_prev.expected_costs(forward(dyn_prev, policy, x0_prev)))

    # Cost under new dynamics and new policy
    actual_new_costs = np.sum(costs.expected_costs(forward(dyn, policy, x0)))

    # Compute step multiplier
    step_mult *= 0.5 * (predicted_new_costs - previous_costs) / (predicted_new_costs - actual_new_costs)

    # Ensure step bounds
    step_mult = np.clip(step_mult, min_step_mult, max_step_mult)

    # Log
    datalogging.log(
        previous_costs=previous_costs,
        predicted_new_costs=predicted_new_costs,
        actual_new_costs=actual_new_costs,
        step_mult=step_mult,
    )

    return step_mult



import numpy as np
import tensorflow as tf
import tensorflow.keras.layers as layers
import tensorflow_probability as tfp
from tqdm import tqdm

from donk.policy.lin_gaussian import LinearGaussianPolicy
from donk.policy.policy import Policy


class Neural_Network_Policy(Policy):
    """Neural network state-action mapping policy."""

    def __init__(self, model, normalize_states: bool = True):
        """Initialize this policy.

        Args:
            model: A `tf.keras.Model`
            normalize_states: Adds a `Normalization` layer directly after the input of the model.

        """
        # Extract dimensions
        self.dX = model.input.shape[-1]
        self.dU = model.output.shape[-1]

        # Add state normalization to model
        if normalize_states:
            self.state_normalization = layers.Normalization(name="normalized_state")

            state = tf.keras.Input((self.dX, ), name="state")
            normalized_state = self.state_normalization(state)
            action = model(normalized_state)

            self.model = tf.keras.Model(inputs=state, outputs=action)
        else:
            self.state_normalization = None
            self.model = model

        # Add metrics to model
        self.model.metric_loss_kl = tf.keras.metrics.Mean("train/loss_kl")
        self.model.metric_loss_reg = tf.keras.metrics.Mean("train/loss_reg")
        self.model.metric_loss = tf.keras.metrics.Mean("train/loss")
        self.model.metric_loss_val = tf.keras.metrics.Mean("val/loss")

    def update(
        self,
        X_train: np.ndarray,
        U_train: np.ndarray,
        prc_train: np.ndarray,
        epochs: int,
        batch_size: int,
        X_val: np.ndarray = None,
        U_val: np.ndarray = None,
        prc_val: np.ndarray = None,
        silent: bool = False
    ):
        """Train the model on new data.

        May supply validation data.

        Args:
            X_train: (N_train, dX) Training states
            U_train: (N_train, dU) Training target actions
            prc_train: (N_train, dU, dU) Training action precisions
            epochs: Number of epochs
            batch_size: Batch size for training
            X_val: (N_val, dX) Validation states (optional)
            U_val: (N_val, dU) Validation target actions (optional)
            prc_val: (N_val, dU, dU) Validation action precisions (optional)
            silent: Whether to disable progress bar
        """
        N_train, _ = X_train.shape

        # Normalize states
        if self.state_normalization is not None:
            self.state_normalization.adapt(X_train)

        # Normalize precision
        prc_scale = np.einsum("ijj->", prc_train) / N_train / self.dU
        prc_train = prc_train / prc_scale
        if prc_val is not None:
            prc_val = prc_val / prc_scale

        # Build dataset
        dataset = tf.data.Dataset.from_tensor_slices(
            (
                X_train.astype(np.float32, copy=False),
                U_train.astype(np.float32, copy=False),
                prc_train.astype(np.float32, copy=False),
            )
        ).shuffle(N_train).batch(batch_size)
        if X_val is not None:
            dataset_test = tf.data.Dataset.from_tensor_slices(
                (
                    X_val.astype(np.float32, copy=False),
                    U_val.astype(np.float32, copy=False),
                    prc_val.astype(np.float32, copy=False),
                )
            ).batch(256)

        @tf.function
        def train_step(model, opt, state, action, precision):
            """Perform one train step."""
            with tf.GradientTape() as tape:
                action_pred = model(state, training=True)

                # KL divergence loss
                #  loss_kl = 1/2 delta_action^T * prc * delta_action
                delta_action = action - action_pred
                loss_kl = tf.reduce_mean(tf.einsum("in,inm,im->i", delta_action, precision, delta_action)) / 2

                # Regularization loss
                loss_reg = tf.reduce_sum(model.losses)

                # Total loss
                loss = loss_kl + loss_reg

            opt.apply_gradients(zip(tape.gradient(loss, model.trainable_variables), model.trainable_variables))

            # Update metrics
            self.model.metric_loss_kl(loss_kl)
            self.model.metric_loss_reg(loss_reg)
            self.model.metric_loss(loss)

        @tf.function
        def validation_step(model, state, action, precision):
            """Perform one validation step."""
            action_pred = model(state, training=False)

            delta_action = action - action_pred
            loss = tf.reduce_mean(tf.einsum("in,inm,im->i", delta_action, precision, delta_action)) / 2
            model.metric_loss_val(loss)

        # Reset optimizer
        opt = tf.keras.optimizers.Adam()

        with tqdm(range(epochs), disable=silent) as pbar:
            for epoch in pbar:
                self.model.reset_metrics()

                # Train batches
                for state, action, precision in dataset:
                    train_step(self.model, opt, state, action, precision)

                # Validation batches
                if X_val is not None:
                    for state, action, precision in dataset_test:
                        validation_step(self.model, state, action, precision)

                # Log metrics
                for metric in self.model.metrics:
                    tf.summary.scalar(metric.name, metric.result(), step=epoch)

                # Update progress bar
                pbar.set_description(
                    f"Train loss: {self.model.metric_loss.result():.6f}" +
                    (f" Val loss: {self.model.metric_loss_val.result():.6f}" if X_val is not None else "")
                )

    def act(self, x: np.ndarray, t: int = None, noise: np.ndarray = None) -> np.ndarray:
        """Decide an action for the given state(s).

        Args:
            x: (..., dX) State(s)
            t: Timestep, ignored

        Returns:
            u: (..., dU) Selected action(s)
        """
        if noise is not None:
            raise NotImplementedError(f"Noise is not supported by {type(self)}")

        u = self.model(x.reshape(-1, self.dX).astype(np.float32, copy=False), training=False).numpy()
        return u.reshape(x.shape[:-1] + (self.dU, ))

    def linearize(self, X: np.ndarray, regularization: float = 1e-6) -> LinearGaussianPolicy:
        """Compute linearization of this policy.

        Args:
            X: (N, T, dX) States

        Returns:
            pol_lin: LinearGaussianPolicy
        """
        N, T, _ = X.shape
        dX, dU = self.dX, self.dU

        x = tf.convert_to_tensor(X.reshape(N * T, dX), dtype=tf.float32)  # Flatten states
        with tf.GradientTape() as tape:
            tape.watch(x)
            U = tf.reshape(self.model(x, training=False), (N, T, dU))
            U_mean = tf.reduce_mean(U, axis=0)
            U_sum = tf.reduce_sum(U_mean, axis=0)  # Sum over timesteps as they are independend, to reduce dimension of du_dx

        # Compute and reshape jacobian
        du_dx = tape.jacobian(U_sum, x)
        du_dx = tf.reshape(du_dx, (dU, N, T, dX))
        du_dx = tf.reduce_mean(du_dx, axis=1)
        du_dx = tf.transpose(du_dx, (1, 0, 2))
        # du_dx.shape = (T, dU, dX)

        # Consutruct policy object
        K = du_dx.numpy()
        k = U_mean.numpy() - np.einsum("tux,tx->tu", K, np.mean(X, axis=0))
        pol_covar = tfp.stats.covariance(U).numpy() + np.eye(self.dU) * regularization  # Use sample covariance

        return LinearGaussianPolicy(K, k, pol_covar)

    def load_weights(self, data_file):
        """Load the network weighs from a file.

        Args:
            data_file: File to load weights from
        """
        self.model.load_weights(data_file)

    def save_weights(self, data_file):
        """Save the network weighs in a file.

        Args:
            data_file: File to store weights to
        """
        self.model.save_weights(data_file)



from __future__ import annotations

from pathlib import Path

import numpy as np
from scipy.stats import multivariate_normal

from donk.dynamics import DynamicsModel
from donk.dynamics.prior import DynamicsPrior
from donk.models import TimeVaryingLinearGaussian
from donk.utils.batched import regularize, symmetrize

multivariate_normal_logpdf = np.vectorize(multivariate_normal.logpdf, signature="(x),(x),(x,x)->()")


class LinearDynamics(DynamicsModel, TimeVaryingLinearGaussian):
    """Stochastic linear dynamics model."""

    def __init__(self, F: np.ndarray, f: np.ndarray, covar: np.ndarray):
        """Initialize this LinearDynamics object.

        Args:
            Fm: (T, dX, dX + dU), Linear term
            fv: (T, dX), Constant term
            dyn_covar: (T, dX, dX), Covariances

        """
        TimeVaryingLinearGaussian.__init__(self, F, f, covar)
        _, self.dX = f.shape
        self.dU = F.shape[2] - self.dX

    F = TimeVaryingLinearGaussian.coefficients
    f = TimeVaryingLinearGaussian.intercept

    def predict(self, x: np.ndarray, u: np.ndarray, t: int = None, noise: np.ndarray = None):
        """Predict the next state.

        Samples next state from the Gaussian distributing given by x' ~ N(Fm_t * [x; u] + fv_t, dyn_covar_t).

        Args:
            x: (dX,) Current state
            u: (dU,) Current action
            t: Current timestep
            noise: (dX,) State noise, may be `None` to sample without noise

        Returns:
            x: (dX,) Next state
        """
        return TimeVaryingLinearGaussian.predict(self, np.concatenate([x, u], axis=-1), t, noise)

    def __str__(self) -> str:
        return f"LinearDynamics[T={self.T}, dX={self.dX}, dU={self.dU}]"

    def log_prob(self, X: np.ndarray, U: np.ndarray) -> float:
        """Compute log probabilities for transitions to occur under this dynamics.

        Args:
            X: (..., T+1, dX) States
            U: (..., T, dU) Actions
        Returns:
            log_prob: (..., T) log transition probabilities
        """
        means = self.predict(X[..., :-1, :], U, t=None)
        return multivariate_normal_logpdf(X[..., 1:, :], means, self.covar)

    def evaluate(self, output_dir: str | Path, X_train: np.ndarray, U_train: np.ndarray, X_test: np.ndarray, U_test: np.ndarray):
        """Create diagnostics and evaluation plots for this dynamics model.

        Args:
            output_dir: Directory to write the plots to
            X_train: (N_train, T+1, dX), Train set states
            U_train: (N_train, T, dU), Train set actions
            X_test: (N_test, T+1, dX), Test set states
            U_test: (N_test, T, dU), Test set actions
        """
        import pandas as pd

        import donk.visualization as vis

        if output_dir is not None:
            output_dir = Path(output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)

        X = np.concatenate([X_train, X_test], axis=0)
        U = np.concatenate([U_train, U_test], axis=0)

        # Compute prediction errors
        N_test = X_test.shape[0]
        prediction = np.empty((N_test, self.T, self.dX))
        for n in range(N_test):
            for t in range(self.T):
                prediction[n, t] = self.predict(X_test[n, t], U_test[n, t], t, noise=None)

        # Create plots
        if output_dir is not None:
            vis.visualize_linear_model(
                output_dir / "parameters.pdf",
                self.Fm,
                self.fv,
                self.dyn_covar,
                x=np.mean(np.c_[X_train[:, :-1], U_train], axis=0),  # Sample mean
            )
            vis.visualize_coefficients(str(output_dir / "coefficients_{:02d}.pdf"), self.Fm)
            vis.visualize_covariance(output_dir / "covariance.pdf", self.dyn_covar.mean(axis=0))
            vis.visualize_prediction(str(output_dir / "prediction_{:02d}.pdf"), prediction, X_test[:, 1:])
            vis.visualize_prediction_error(output_dir / "error.pdf", prediction, X_test[:, 1:])
            vis.visualize_predictor_target_correlation(
                output_dir / "state_correlation.pdf",
                X=np.concatenate([X[:, :-1], U], axis=-1).reshape(-1, self.dX + self.dU),
                Y=X[:, 1:].reshape(-1, self.dX),
                xlabel="$xu_t$",
                ylabel="$x_{t+1}$",
            )
            vis.visualize_predictor_target_scatter(
                output_dir / "state_correlation_scatter.png", X=np.concatenate([X[:, :-1], U], axis=-1), Y=X[:, 1:]
            )

        # Gather statistics
        statistics = pd.DataFrame(
            [
                ("T", self.T),
                ("dX", self.dX),
                ("dU", self.dU),
                ("coefficients_variance", np.var(self.Fm, axis=0).mean()),
                ("prediction_error", np.mean((prediction - X_test[:, 1:])**2)),
            ],
            columns=["metric", "score"],
        )
        if output_dir is not None:
            statistics.to_csv(output_dir / "statistics.csv", index=False)

        return statistics


def fit_lr(X: np.ndarray, U: np.ndarray, prior: DynamicsPrior = None, regularization=1e-6) -> LinearDynamics:
    """Fit dynamics with least squares linear regression.

    Args:
        X: (N, T+1, dX), States
        U: (N, T, dU), Actions
        prior: DynamicsPrior to be used. May be `None` to fit without prior.
        regularization: Eigenvalues of the joint distribution variance are lifted to this value. Ensures matrix is not singular.
    """
    N, _, dX = X.shape
    _, T, dU = U.shape
    dXU = dX + dU

    if N <= 1:
        raise ValueError(f"Cannot fit dynamics to {N} sample(s)")

    Fm = np.empty([T, dX, dX + dU])
    fv = np.empty([T, dX])
    dyn_covar = np.empty([T, dX, dX])

    # Perform regression for all time steps
    for t in range(T):
        xux = np.c_[X[:, t], U[:, t], X[:, t + 1]]
        empmu = np.mean(xux, axis=0)
        empsig = (xux - empmu).T @ (xux - empmu) / N

        if prior is None:
            mu = empmu
            sigma = empsig
        else:
            prior_dist = prior.eval(xux)
            posterior_dist = prior_dist.posterior(empmu, empsig, N)

            # Override the size of the prior
            # TODO More elegant solution
            posterior_dist.N_mean = 1
            posterior_dist.N_covar = -(dX + dU + dX)

            mu = empmu  # Instead of using the estimation, suggested by Finn et al. in https://github.com/cbfinn/gps
            sigma = posterior_dist.map_covar()

        # Apply regularization to ensure non-sigularity
        min_eigval = np.min(np.linalg.eigvalsh(sigma[:dXU, :dXU]))
        if min_eigval < regularization:
            # Lift eigenvalues lowest eigenvalue to regularization
            regularize(sigma[:dXU, :dXU], regularization - min_eigval)

        # Condition on x_t, u_t
        Fm[t] = np.linalg.solve(sigma[:dXU, :dXU], sigma[:dXU, dXU:]).T
        fv[t] = mu[dXU:] - Fm[t] @ (mu[:dXU])
        dyn_covar[t] = sigma[dXU:, dXU:] - Fm[t] @ sigma[:dXU, :dXU] @ Fm[t].T

    symmetrize(dyn_covar)
    return LinearDynamics(Fm, fv, dyn_covar)



from abc import ABC, abstractmethod

import numpy as np

import donk


class CostFunction(ABC):
    """Base class for cost functions.

    Defines a `quadratic_approximation` method.
    """

    @abstractmethod
    def quadratic_approximation(self, X: np.ndarray, U: np.ndarray) -> "donk.costs.quadratic_costs.QuadraticCosts":
        """Compute a quadratic approximation (2nd order Taylor) at the given trajectory.

        Args:
            X: (T+1, dX), states
            U: (T, dX), actions
        """

    @abstractmethod
    def compute_costs(self, X: np.ndarray, U: np.ndarray) -> np.ndarray:
        """Evaluate costs for trajectories.

        Args:
            X: (..., T+1, dX), states
            U: (..., T, dU), actions

        Returns:
            costs: (..., T+1), Costs at each time step
        """



from abc import ABC, abstractmethod

import numpy as np


class Policy(ABC):
    """Base class for policies.

    Defines an `act` method.
    """

    @abstractmethod
    def act(self, x: np.ndarray, t: int = None, noise: np.ndarray = None) -> np.ndarray:
        """Decide an action for the given state at the current timestep.

        For time-varying policies `t` is required.

        Args:
            x: (dX,) Current state
            t: Current timestep, may be `None`
            noise: (dU,) Action noise, may be `None` to sample without noise

        Returns:
            u: (dU,) Selected action
        """



from pathlib import Path

import numpy as np

from donk.policy import LinearGaussianPolicy


def load_state_controller_dataset(dataset, itr):
    """Loads a state_controller dataset.

    Args:
        dataset: Id of the dataset
        itr: Iteration to return

    Returns:
        X: (N, T, dX) Real states
        pol: Fitted linear policy
        X_mean: (T, dX) Mean of state distribution
        X_covar: (T, dX, dX) Covariance of state distribution
    """
    file = Path(f"tests/data/state_controller_{dataset:02d}.npz")
    if not file.is_file():
        raise ValueError(f"There is no dataset 'state_controller_{dataset:02d}.npz'")

    with np.load(file) as data:
        if itr not in range(len(data['X'])):
            raise ValueError(f"Invalid iteration {itr}")

        X = data["X"][itr]
        pol = LinearGaussianPolicy(K=data["K"][itr], k=data["k"][itr], pol_covar=data["pol_covar"][itr])
        X_mean = data["X_mean"][itr]
        X_covar = data["X_covar"][itr]

    return X, pol, X_mean, X_covar



from __future__ import annotations

import dbm.dumb
import pickle
from typing import List

from donk.datalogging.datalogging import DataLogger


class ShelveDataLogger(DataLogger):
    """Log data as single shelve file."""

    def __init__(self, file) -> None:
        super().__init__()

        self.file = file

    def log(self, key: List[str], data) -> None:
        """Log one data.

        Args:
            key: Path-like, hierarchical name associated with the data
            data: Data to log, should be python primitives
        """
        # Store data
        self.__data["/".join(key)] = pickle.dumps(data)

    def flush(self) -> None:
        """Write data to the file."""
        self.__data.sync()

    def __enter__(self) -> ShelveDataLogger:
        # Open database
        self.__data = dbm.dumb.open(self.file, "c")

        return super().__enter__()

    def __exit__(self, type, value, traceback):
        super().__exit__(type, value, traceback)

        # Close shelve
        self.__data.close()



import numpy as np
from scipy.ndimage import gaussian_filter1d


def smooth_noise(noise: np.ndarray, kernel_std: float) -> np.ndarray:
    """Apply a Gaussian filter to smooth noise.

    Maintains mean and variance of input.

    Args:
        noise: (T, dU), input noise array
        kernel_std: Size of the Gaussian Kernel

    Returns:
        smoothed: (T, dU), smoothed noise array
    """
    _, dU = noise.shape

    # Apply Gaussian filter
    smoothed = np.empty_like(noise)
    for i in range(dU):
        smoothed[:, i] = gaussian_filter1d(noise[:, i], kernel_std, mode="nearest")

    # Renorm
    smoothed *= np.std(noise, axis=0) / np.std(smoothed, axis=0)
    smoothed += np.mean(noise, axis=0) - np.mean(smoothed, axis=0)
    return smoothed



import numpy as np

from donk.utils.batched import batched_cholesky, batched_inv_spd


class TimeVaryingLinearGaussian:
    """Time-varying linear model with Gaussian noise.

    `p(y|t,x) ~ N(coeffcients_t * x + intercept_t, covar_t)`
    """

    def __init__(self, coefficients: np.ndarray, intercept: np.ndarray, covar: np.ndarray = None, inv_covar: np.ndarray = None) -> None:
        """Initialize this TimeVaryingLinearGaussian object.

        Must provide either covariance or precision, or both.

        Args:
            coefficients: (T, dY, dX), Linear term
            intercept: (T, dY), Constant term
            covar: (T, dY, dY), Covariances
            inv_covar: (T, dY, dY), Inverse covariances, precision.
        """
        if covar is None and inv_covar is None:
            raise ValueError("Must provide covar or inv_covar.")

        self.T, dY, dX = coefficients.shape

        # Check shapes
        assert coefficients.shape == (self.T, dY, dX), f"{coefficients.shape} != {(self.T, dY, dX )}"
        assert intercept.shape == (self.T, dY), f"{intercept.shape} != {(self.T, dY)}"
        if covar is not None:
            assert covar.shape == (self.T, dY, dY), f"{covar.shape} != {(self.T, dY, dY)}"
        if inv_covar is not None:
            assert inv_covar.shape == (self.T, dY, dY), f"{inv_covar.shape} != {(self.T,dY, dY)}"

        self._coefficients = coefficients
        self._intercept = intercept
        self._covar = covar
        self._chol_covar = None
        self._inv_covar = inv_covar

    @property
    def coefficients(self) -> np.ndarray:
        """Linear coefficients of this model.

        Returns:
            coefficients: (T, dY, dX)
        """
        return self._coefficients

    @property
    def intercept(self) -> np.ndarray:
        """Constant intercepts of this model.

        Returns:
            intercept: (T, dY)
        """
        return self._intercept

    @property
    def covar(self) -> np.ndarray:
        """Covariances of this model.

        Computed lazily, if required.

        Returns:
            covar: (T, dY, dY)
        """
        if self._covar is None:
            self._covar = batched_inv_spd(batched_cholesky(self._inv_covar))
        return self._covar

    @property
    def chol_covar(self) -> np.ndarray:
        """Cholesky decoposition of the covariances of this model.

        Computed lazily, if required.

        Returns:
            chol_covar: (T, dY, dY)
        """
        if self._chol_covar is None:
            self._chol_covar = batched_cholesky(self.covar)
        return self._chol_covar

    @property
    def inv_covar(self) -> np.ndarray:
        """Inverse covariances/precisions of this model.

        Computed lazily, if required.

        Returns:
            inv_covar: (T, dY, dY)
        """
        if self._inv_covar is None:
            self._inv_covar = batched_inv_spd(self.chol_covar)
        return self._inv_covar

    def predict(self, X: np.ndarray, t: int = None, noise: np.ndarray = None) -> np.ndarray:
        """Sample preditions from this model.

        Samples the Gaussian distributing given by `p(y|t,x) ~ N(coeffcients_t * x + intercept_t, covar_t)`.

        Can either sample a specific timestep, or from all.

        Noise can be omitted to compute the output expectations.

        Args:
            X: (..., dX) or (..., T, dX) Inputs
            t: Supply to sample from a specific timestept, optional
            noise: (..., dY,) Output noise, optional

        Returns:
            Y: (..., dY) or  (..., T, dY) Outputs
        """
        if t is not None:
            Y = np.einsum("ux,...x->...u", self.coefficients[t], X) + self.intercept[t]
            if noise is not None:
                Y += np.einsum("ij,...j->...i", self.chol_covar[t], noise)
        else:
            dX = self.coefficients.shape[-1]
            if X.shape[-2:] != (self.T, dX):
                raise ValueError(f"Input must have shape (..., {self.T}, {dX}), not {X.shape}")
            Y = np.einsum("tyx,...tx->...ty", self.coefficients, X) + self.intercept
            if noise is not None:
                Y += np.einsum("tij,...tj->...ti", self.chol_covar, noise)
        return Y



import numpy as np

from donk.models import TimeVaryingLinearGaussian
from donk.policy.policy import Policy


class LinearGaussianPolicy(Policy, TimeVaryingLinearGaussian):
    """Time-varying linear Gaussian policy.

    `p(u|t,x) ~ N(K_t * x + k_t, covar_t)`
    """

    def __init__(self, K: np.ndarray, k: np.ndarray, covar: np.ndarray = None, inv_covar: np.ndarray = None):
        """Initialize this LinearGaussianPolicy object.

        Must provide either covariance or precision, or both.

        Args:
            K: (T, dU, dX), Linear term
            k: (T, dU), Constant term
            covar: (T, dU, dU), Covariances
            inv_covar: (T, dU, dU), Inverse covariances, also called precision.
        """
        TimeVaryingLinearGaussian.__init__(self, K, k, covar, inv_covar)
        _, self.dU, self.dX = K.shape

    K = TimeVaryingLinearGaussian.coefficients
    k = TimeVaryingLinearGaussian.intercept

    def act(self, x: np.ndarray, t: int, noise: np.ndarray = None) -> np.ndarray:
        """Decide an action for the given state at the current timestep.

        Uses the TVLG model to predict the next action.

        Args:
            x: (dX,) Current state
            t: Current timestep
            noise: (dU,) Action noise, may be `None` to sample without noise

        Returns:
            u: (dU,) Selected action
        """
        return self.predict(x, t, noise)

    def __str__(self) -> str:
        return f"LinearGaussianPolicy[T={self.T}, dX={self.dX}, dU={self.dU}]"



from donk.costs.cost_function import CostFunction
from donk.costs.cost_function_symbolic import MultipartSymbolicCostFunction, SymbolicCostFunction
from donk.costs.losses import loss_combined, loss_l1, loss_l2, loss_log_cosh
from donk.costs.quadratic_costs import QuadraticCosts

__all__ = [
    "loss_l2",
    "loss_l1",
    "loss_log_cosh",
    "loss_combined",
    "QuadraticCosts",
    "CostFunction",
    "SymbolicCostFunction",
    "MultipartSymbolicCostFunction",
]



from pathlib import Path
from typing import List

import matplotlib.pyplot as plt
import numpy as np


def visualize_costs(
    output_file: Path,
    costs: List[np.ndarray],
    cost_labels: List[str],
    include_total: bool = True,
):
    """Plots mutiple cost curves.

    Args:
        output_file: File to write plot to. Use `None` to show plot.
        costs: [(..., T+1)] Costs
        cost_labels: Names for the costs
        include_total: Sum up all costs and show a total
    """
    T = costs[0].shape[-1] - 1

    if include_total:
        costs.append(np.sum(costs, axis=0))
        cost_labels.append("total")

    for name, cost in zip(cost_labels, costs):
        cost = cost.reshape(-1, T + 1)
        l = plt.plot(np.mean(cost, axis=0), label=name)[0]

        plt.fill_between(
            np.arange(T + 1),
            np.min(cost, axis=0),
            np.max(cost, axis=0),
            facecolor=l.get_color(),
            alpha=0.25,
            interpolate=True,
        )

    plt.legend()
    plt.xlabel("$t$")
    plt.ylabel("costs")

    # Export
    plt.tight_layout()
    if output_file is not None:
        plt.savefig(output_file)
    else:
        plt.show()
    plt.close()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List


class DataLogger(ABC):
    """Base class for data loggers."""

    @abstractmethod
    def log(self, key: List[str], data) -> None:
        """Log one data.

        Args:
            key: Path-like, hierarchical name associated with the data
            data: Data to log, should be python primitives
        """

    def as_default(self) -> DataLogger:
        """Make the the active data logger."""
        global _data_logger
        _data_logger = self
        return self

    def flush(self) -> None:
        """Flush data, if applicable."""
        pass

    def __enter__(self) -> DataLogger:
        global _data_logger

        # Store previous logger
        self.stored_logger = _data_logger

        return self.as_default()

    def __exit__(self, type, value, traceback):
        self.flush()

        # Restore stored logger
        self.stored_logger.as_default()


class DummyDataLogger(DataLogger):
    """Logs directly to the void."""

    def log(self, key: list[str], data):
        """Ignore one data.

        Args:
            key: Path-like, hierarchical name associated with the data
            data: Data to log, should be python primitives
        """
        pass  # Do nothing


class Context():

    def __init__(self, name: str):
        self.name = name

    def __enter__(self):
        global _data_logger_context

        # Store previous context
        self.stored_context = _data_logger_context

        # Append new context
        _data_logger_context = _data_logger_context + [self.name]

        return self

    def __exit__(self, type, value, traceback):
        global _data_logger_context

        # Restore context
        _data_logger_context = self.stored_context


# Global state for the active logger
_data_logger: DataLogger = DummyDataLogger()
_data_logger_context = []


def log(**kwargs):
    """Logs one or multiple pieces of data.

    Args:
        key: Name for the piece of data
        data: Data to log
    """
    for key, data in kwargs.items():
        _data_logger.log(_data_logger_context + [key], data)



from donk.datalogging.datalogging import Context, DataLogger, log
from donk.datalogging.shelve import ShelveDataLogger

__all__ = [
    "DataLogger",
    "log",
    "Context",
    "ShelveDataLogger",
]



from donk.dynamics.prior.prior import DynamicsPrior, NormalInverseWishart
from donk.dynamics.prior.prior_gmm import GMMPrior

__all__ = [
    "DynamicsPrior",
    "GMMPrior",
    "NormalInverseWishart",
]



"""Visualization tool for linear models."""
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.ticker import MaxNLocator

from donk.dynamics import LinearDynamics
from donk.policy import LinearGaussianPolicy


def visualize_linear_model(
    output_file: Path,
    coeff,
    intercept,
    cov,
    x,
    y=None,
    N=100,
    coeff_label="coefficients",
    intercept_label="intercept",
    cov_label="covariance",
    y_label="prediction",
    time_label="$t$",
    export_data=True,
):
    """Creates a figure visualizing a timeseries of linear Gausian models.

    Args:
        output_file: File to write the plot to.
        coeff: Linear coefficients. Shape: (T, dY, dX)
        intercept: Constants. Shape: (T, dY)
        cov: Covariances. Shape: (T, dY, dY)
        x: Shape (T, dX)
        y: Optional. Shape (T, dY)
        N: Number of random samples drawn to visualize variance.
    """
    fig = plt.figure(figsize=(16, 12))

    T, dY, dX = coeff.shape

    # Check shapes
    assert intercept.shape == (T, dY), f"{intercept.shape} != {(T, dY, dX)}"
    assert cov.shape == (T, dY, dY), f"{cov.shape} != {(T, dY, dY)}"
    assert x.shape == (T, dX), f"{x.shape} != {(T, dX)}"
    if y is not None:
        assert y.shape == (T, dY), f"{y.shape} != {(T, dY)}"

    # Intercept
    ax1 = fig.add_subplot(221)
    ax1.set_ylabel(intercept_label)
    ax1.set_xlabel(time_label)
    ax1.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax1.grid(linestyle=":")
    for dim in range(dY):
        line = ax1.plot(np.arange(T), intercept[:, dim], linewidth=1)[0]

    # Coefficients
    ax2 = fig.add_subplot(222, sharex=ax1)
    ax2.set_ylabel(coeff_label)
    ax2.set_xlabel(time_label)
    ax2.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax2.grid(linestyle=":")
    for dim1 in range(dY):
        for dim2 in range(dX):
            line = ax2.plot(np.arange(T, dtype=int), coeff[:, dim1, dim2], linewidth=1)[0]

    # Covariance
    ax3 = fig.add_subplot(223, sharex=ax1)
    ax3.set_ylabel(cov_label)
    ax3.set_xlabel(time_label)
    ax3.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax3.grid(linestyle=":")
    for dim1 in range(dY):
        for dim2 in range(dY):
            line = ax3.plot(np.arange(T), cov[:, dim1, dim2], linewidth=1)[0]

    # Prediction
    y_ = np.empty((N, T, dY))  # Approx y using the model
    for t in range(T):
        mu = coeff[t] @ x[t] + intercept[t]
        y_[:, t] = np.random.multivariate_normal(mean=mu, cov=cov[t], size=N)
    y_mean = np.mean(y_, axis=0)
    y_std = np.std(y_, axis=0)

    ax4 = fig.add_subplot(224, sharex=ax1)
    ax4.set_ylabel(y_label)
    ax4.set_xlabel(time_label)
    ax4.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax4.grid(linestyle=":")
    for dim in range(dY):
        line = ax4.plot(np.arange(T), y_mean[:, dim], linewidth=1)[0]
        c = line.get_color()
        if y is not None:
            ax4.plot(np.arange(T), y[:, dim], ":", color=c)
        ax4.fill_between(
            np.arange(T),
            y_mean[:, dim] - y_std[:, dim],
            y_mean[:, dim] + y_std[:, dim],
            facecolor=c,
            alpha=0.25,
            interpolate=True,
        )

    # Export
    plt.tight_layout()
    if output_file is not None:
        fig.savefig(output_file)
        if export_data:
            data_file = output_file.parent / output_file.stem
            np.savez_compressed(data_file, coeff=coeff, intercept=intercept, cov=cov, y=y, y_mean=y_mean, y_std=y_std)
    else:
        plt.show()
    plt.close(fig)


def visualize_linear_dynamics_model(output_file: Path, dyn: LinearDynamics, X: np.ndarray, U: np.ndarray, **kwargs):
    """Creates a figure visualizing a TVLG dynamics model.

    Args:
        output_file: File to write the plot to.
        dyn: Linear dynamics model
        X: (N, T+1, dX) States
        U: (N, T+1, dU) Actions
        kwargs: Passed to `visualize_linear_model`
    """
    visualize_linear_model(
        output_file,
        dyn.F,
        dyn.f,
        dyn.covar,
        np.mean(np.concatenate([X[:, :-1], U], axis=-1), axis=0),
        np.mean(X[:, 1:], axis=0),
        coeff_label=r"$\mathbf{F}_t$",
        intercept_label=r"$\mathbf{f}_t$",
        cov_label=r"$\mathbf{\Sigma}_t^{dyn}$",
        y_label=r"$\mathbf{x}_{t+1}$",
        **kwargs
    )


def visualize_linear_policy(output_file: Path, pol: LinearGaussianPolicy, X: np.ndarray, **kwargs):
    """Creates a figure visualizing a TVLG policy.

    Args:
        output_file: File to write the plot to.
        dyn: Linear dynamics model
        X: (N, T+1, dX) States
        kwargs: Passed to `visualize_linear_model`
    """
    visualize_linear_model(
        output_file,
        pol.K,
        pol.k,
        pol.covar,
        np.mean(X[:, :-1], axis=0),
        coeff_label=r"$\mathbf{K}_t$",
        intercept_label=r"$\mathbf{k}_t$",
        cov_label=r"$\mathbf{\Sigma}_t^{pol}$",
        y_label=r"$\mathbf{u}_t$",
        **kwargs
    )


def visualize_coefficients(output_file_pattern, coeff):
    """Visualize coefficietns of a linear model.

    Args:
        output_file_pattern: Pattern for files to write the plots to.
        coeff: shape (T, dY, dX), linear coefficients
    """

    T, dY, dX = coeff.shape

    for y in range(dY):
        df = pd.DataFrame(
            ((t, x, coeff[t, y, x]) for t in range(T) for x in range(dX)),
            columns=["t", "x", "coeff"],
        )

        sns.boxplot(data=df, x="x", y="coeff")
        plt.ylabel(f"$coeff_{{{y},x}}$")

        plt.tight_layout()
        if output_file_pattern is not None:
            plt.savefig(output_file_pattern.format(y))
        else:
            plt.show()
        plt.close()


def visualize_covariance(output_file, covar):
    """Visualize the covariance matrix of a linear model.

    Args:
        output_file: File to write the plot to.
        covar: shape (dX, dX), covariance matrix
    """
    sns.heatmap(
        covar,
        mask=np.triu(np.ones_like(covar, dtype=bool), 1),  # Mask upper triangle
    )

    plt.tight_layout()
    if output_file is not None:
        plt.savefig(output_file)
    else:
        plt.show()
    plt.close()


def visualize_prediction(output_file_pattern, prediction, truth):
    """Visualize the prediction errors of a linear model.

    Args:
        output_file_pattern: Pattern for files to write the plots to.
        prediction: shape (N, T, dY), model predictions
        truth: shape (N, T, dY), target values
    """
    N, T, dY = prediction.shape

    for y in range(dY):
        if N <= 4:
            for n in range(N):
                plt.subplot(2, 2, 1 + n)
                plt.plot(prediction[n, :, y], c="C0", linewidth=1, label="prediction" if n + 1 == N else None)
                plt.plot(truth[n, :, y], c="C1", linewidth=1, label="truth" if n + 1 == N else None)
            plt.legend()
        else:
            plt.plot(prediction[:, :, y].mean(axis=0), c="C0", linewidth=1)
            plt.fill_between(
                np.arange(T),
                prediction[:, :, y].min(axis=0),
                prediction[:, :, y].max(axis=0),
                facecolor="C0",
                alpha=0.25,
                interpolate=True
            )
            plt.plot(truth[:, :, y].mean(axis=0), c="C1", linewidth=1)
            plt.fill_between(
                np.arange(T),
                truth[:, :, y].min(axis=0),
                truth[:, :, y].max(axis=0),
                facecolor="C1",
                alpha=0.25,
                interpolate=True,
            )

        plt.tight_layout()
        if output_file_pattern is not None:
            plt.savefig(output_file_pattern.format(y))
        else:
            plt.show()
        plt.close()


def visualize_prediction_error(output_file, predictions, targets):
    """Visualize the prediction errors of a linear model.

    Args:
        output_file: File to write the plot to.
        prediction: shape (N, T, dX), model predictions
        target: shape (N, T, dX), targets
    """
    from sklearn.metrics import explained_variance_score, r2_score

    errors = (predictions - targets)**2
    mse = np.mean(errors)
    N, T, dX = errors.shape
    r2 = r2_score(targets.reshape(N * T, dX), predictions.reshape(N * T, dX))
    explained_variance = explained_variance_score(targets.reshape(N * T, dX), predictions.reshape(N * T, dX))

    plt.figure(figsize=(12, 8))

    # Error over time
    plt.subplot(2, 2, 1)
    sns.lineplot(
        x=np.tile(np.arange(T), (N, 1)).flatten(),
        y=np.mean(errors, axis=2).flatten(),
    )
    plt.axhline(mse, color="black", linewidth=1, linestyle="dashed")
    plt.xlabel("$t$")
    plt.ylabel("$err$")

    # Error over states
    plt.subplot(2, 2, 2)
    sns.barplot(
        x=np.tile(np.arange(dX), (N, T, 1)).flatten(),
        y=errors.flatten(),
    )
    plt.axhline(mse, color="black", linewidth=1, linestyle="dashed")
    plt.xlabel("state")
    plt.ylabel("$err$")

    # Error over samples
    plt.subplot(2, 2, 3)
    sns.barplot(
        x=np.tile(np.arange(N), (1, T)).flatten(),
        y=np.mean(errors, axis=2).flatten(),
    )
    plt.axhline(mse, color="black", linewidth=1, linestyle="dashed")
    plt.xlabel("sample")
    plt.ylabel("$err$")

    plt.subplot(2, 2, 4)
    plt.table(
        [
            ["MSE", f"{mse:.04f}"],
            ["$R^2$", f"{r2:.04f}"],
            ["Explained variance", f"{explained_variance:.04f}"],
        ],
        loc="center",
        cellLoc="left",
        colWidths=[0.25, 0.25],
        edges="open",
    )
    plt.axis("off")

    plt.tight_layout()
    if output_file is not None:
        plt.savefig(output_file)
    else:
        plt.show()
    plt.close()


def visualize_predictor_target_correlation(output_file, X, Y, xlabel="$x$", ylabel="$y$"):
    """Visualize the correlation between predictors and targets.

    Args:
        output_file: File to write the plot to.
        X: shape (N, dX), predictors
        Y: shape (N, dY), targets
    """
    dX, dY = X.shape[-1], Y.shape[-1]
    corr = np.empty((dY, dX))
    for i in range(dX):
        for j in range(dY):
            corr[j, i] = np.corrcoef(X[:, i], Y[:, j])[0, 1]  # TODO can be calculated more efficiently
    sns.heatmap(corr, cmap="bwr", vmin=-1, vmax=1)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title("Correlation")

    plt.tight_layout()
    if output_file is not None:
        plt.savefig(output_file)
    else:
        plt.show()
    plt.close()


def visualize_predictor_target_scatter(output_file, X, Y, cmap=plt.cm.plasma):
    """Visualize the correlation between predictors and targets.

    Args:
        output_file: File to write the plot to.
        X: shape (N, T, dX), predictors
        Y: shape (N, T, dY), targets
    """
    N, T, dX = X.shape
    dY = Y.shape[-1]

    _, axes = plt.subplots(dY, dX, squeeze=False, figsize=(dX * 2, dY * 2))
    colors = cmap(np.tile(np.linspace(0, 1, T), (N, 1)).flatten())
    for y in range(dY):  # Rows
        for x in range(dX):  # Columns
            ax = axes[y][x]
            ax.scatter(X[:, :, x].flatten(), Y[:, :, y].flatten(), c=colors, alpha=0.5)
            ax.axis("equal")
            ax.tick_params(axis="both", which="both", left=False, bottom=False, labelleft=False, labelbottom=False)

    plt.tight_layout()
    if output_file is not None:
        plt.savefig(output_file, dpi=300)
    else:
        plt.show()
    plt.close()



from pathlib import Path

import matplotlib.pyplot as plt
import matplotlib.transforms as mtransforms
import pandas as pd


def visualize_iLQR(output_file: Path, results, kl_step: float, opt_result=None, cost_color="C0", constraint_color="C1"):

    results = pd.DataFrame(results).sort_values("eta")

    fig, ax1 = plt.subplots()
    plt.xlabel(r"$\eta$")
    plt.xscale("log")
    plt.ylabel("costs")

    # Cost plots
    plt.plot(results["eta"], results["expected_costs"], color=cost_color, label="costs")
    if opt_result is not None:
        plt.scatter([opt_result.eta], [opt_result.expected_costs], color=cost_color, label="minimal constrained costs")

    ax2 = ax1.twinx()  # Second y-axis
    plt.ylabel("constraint violation")
    plt.yscale("log")

    # Constraint plots
    ax2.plot(results["eta"], results["kl_div"], label="constraint violation", color=constraint_color)
    ax2.axhline(kl_step, color=constraint_color, linewidth=1, linestyle="dashed", label="constraint threshold")
    ax2.fill_between(
        results["eta"],
        0,
        1,
        where=results["kl_div"] <= kl_step,
        facecolor=constraint_color,
        alpha=0.1,
        transform=mtransforms.blended_transform_factory(ax2.transData, ax2.transAxes),
        label="constraint fulfilled",
    )

    # Figure legend to include both axes
    fig.legend(loc="lower right", bbox_to_anchor=(1, 0), bbox_transform=ax1.transAxes)

    plt.tight_layout()
    plt.savefig(output_file)
    plt.close()

    # Write data to csv
    results.to_csv(
        output_file.parent / (output_file.stem + ".csv"),
        columns=["eta", "kl_div", "expected_costs"],
        index=False,
    )



from donk.samples.pool import TransitionPool
from donk.samples.sampler import Sampler
from donk.samples.sampler_gym import GymSampler
from donk.samples.state_dist import StateDistribution
from donk.samples.traj_dist import TrajectoryDistribution

__all__ = [
    "TransitionPool",
    "StateDistribution",
    "TrajectoryDistribution",
    "Sampler",
    "GymSampler",
]



from donk.dynamics.dynamics import DynamicsModel
from donk.dynamics.linear_dynamics import LinearDynamics
from donk.dynamics.utils import to_transitions

__all__ = [
    "DynamicsModel",
    "LinearDynamics",
    "to_transitions",
]



from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np


@dataclass
class NormalInverseWishart:
    """A normal-inverse Wishart distribution."""
    mu0: np.ndarray
    Phi: np.ndarray
    N_mean: float
    N_covar: float

    def posterior(self, emp_mean: np.ndarray, emp_covar: np.ndarray, N: float):
        """Compute prior distibution after observing samples with the given sample distribution.

        Args:
            emp_mean: (d,) Sample mean
            emp_covar: (d, d) Sample covariance
            N: Number of samples. May be fractional

        See:
            Gelman et al. Bayesian Data Analysis, chapter 3.6
        """
        return NormalInverseWishart(
            mu0=(self.N_mean * self.mu0 + N * emp_mean) / (self.N_mean + N),
            Phi=self.Phi + N * emp_covar + self.N_mean * N /
            (self.N_mean + N) * np.einsum("i,j->ij", emp_mean - self.mu0, emp_mean - self.mu0),
            N_mean=self.N_mean + N,
            N_covar=self.N_covar + N,
        )

    def map_mean(self) -> np.ndarray:
        """Maximum aposteriory estimate for the mean."""
        return self.mu0.copy()

    def map_covar(self) -> np.ndarray:
        """Maximum aposteriory estimate for the covariance."""
        d = self.mu0.shape[-1]
        return self.Phi / (self.N_covar + d + 1)

    @staticmethod
    def non_informative_prior(d):
        """Create a non-onformative prior distribution.

        Args:
            d: Dimension
        """
        return NormalInverseWishart(
            mu0=np.zeros((d, )),
            Phi=np.zeros((d, d)),
            N_mean=0,
            N_covar=-(1 + d),
        )


class DynamicsPrior(ABC):
    """A prior distribution for linear dynamics.

    These priors are given as normal-inverse-Wishart distributions.
    """

    @abstractmethod
    def update(self, XUX: np.ndarray) -> None:
        """Update the prior.

        Args:
            XUX: (N, dX+dU+dX), transitions
        """

    @abstractmethod
    def eval(self, XUX: np.ndarray) -> NormalInverseWishart:
        """Evaluate the prior for the given transitions.

        Args:
            XUX: (N, dX+dU+dX), transitions
        """



"""Visualization tool for state spaces."""
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


def visualize_correlation(output_file, X):
    """Visualize the correlation between states.

    Args:
        output_file: File to write the plot to.
        X: shape (N, dX), states
    """
    corr = np.corrcoef(X, rowvar=False)
    sns.heatmap(
        corr,
        cmap="bwr",
        vmin=-1,
        vmax=1,
        mask=np.triu(np.ones_like(corr, dtype=bool), 1),  # Mask upper triangle
    )
    plt.xlabel("$x$")
    plt.ylabel("$x$")
    plt.title("Correlation")

    plt.tight_layout()
    if output_file is not None:
        plt.savefig(output_file)
    else:
        plt.show()
    plt.close()



import numpy as np


def to_transitions(X: np.ndarray, U: np.ndarray) -> np.ndarray:
    """Convert state-action sequences to state-action-state tuples.

    Args:
        X: (N, T+1, dX), States
        U: (N, T, dU), Actions

    Returns:
        XUX: (N*T, dX+dU+dX)
    """
    N, _, dX = X.shape
    _, T, dU = U.shape
    transitions = np.c_[X[:, :-1], U, X[:, 1:]].reshape(N * T, dX + dU + dX)
    return transitions



import numpy as np


def loss_combined(x, losses):
    """Evaluates and sums up multiple loss functions.

    Args:
        x: (T, dX) states, actual values.
        losses: List of tuples (`loss`, `kwargs`)
            `loss`: loss function to evaluate.
            `kwargs`: Addional arguments passed to the loss function (optional).

    """
    if len(losses) < 1:
        raise ValueError("loss_combined requred at least one loss function to sum up.")

    l, lx, lxx = losses[0][0](x, **losses[0][1])
    for loss in losses[1:]:
        l_, lx_, lxx_ = loss[0](x, **loss[1])
        l += l_
        lx += lx_
        lxx += lxx_

    return l, lx, lxx


def loss_l2(x, t, w):
    """Evaluate and compute derivatives for l2 norm penalty.

    loss = sum(0.5 * (x - t)^2 * w)

    Args:
        x: (T, dX) states, actual values.
        t: (T, dX) targets, expected values.
        w: (T, dX) weights, scale error of each feature at each timestep.

    Returns:
        l: (T,) cost at each timestep.
        lx: (T, D) first order derivative.
        lxx: (T, D, D) second order derivative.

    """
    # Get trajectory length.
    _, dX = x.shape

    d = x - t  # Error

    # Total cost
    # l = sum(0.5 * (x - t)^2 * w)
    l = 0.5 * np.sum(d**2 * w, axis=1)

    # First order derivative
    # lx = (x - t) * w
    lx = d * w

    # Second order derivative
    # lxx = w
    lxx = np.einsum("ij,jk->ijk", w, np.eye(dX))

    return l, lx, lxx


def loss_l1(x, t, w, alpha):
    """Evaluate and compute derivatives for l2 norm penalty.

    loss = sum(sqrt((x - t)^2 + alpha) * w)

    Args:
        x: (T, dX) states, actual values.
        t: (T, dX) targets, expected values.
        w: (T, dX) weights, scale error of each feature at each timestep.

    Returns:
        l: (T,) cost at each timestep.
        lx: (T, D) first order derivative.
        lxx: (T, D, D) second order derivative.

    """
    # Get trajectory length.
    _, dX = x.shape

    d = x - t  # Error
    abs_d = np.sqrt(alpha + d**2)

    # Total cost
    # l = sum(w * sqrt((x - t)^2 + alpha))
    l = np.sum(w * abs_d, axis=1)

    # First order derivative
    # lx = w * (x - t) / sqrt((x - t)^2 + alpha) * w
    lx = w * (x - t) / abs_d

    # Second order derivative
    # lxx = w * alpha / (((x-t)^2 + alpha)^(3/2))
    lxx = np.einsum("ij,jk->ijk", w * alpha / abs_d**3, np.eye(dX))

    return l, lx, lxx


def loss_log_cosh(x, t, w):
    """Evaluate and compute derivatives for log-cosh loss.

    loss = sum(log(cosh(x - t)) * w)

    Args:
        x: (T, dX) states, actual values.
        t: (T, dX) targets, expected values.
        w: (T, dX) weights, scale error of each feature at each timestep.

    Returns:
        l: (T,) cost at each timestep.
        lx: (T, D) first order derivative.
        lxx: (T, D, D) second order derivative.

    """
    # Get trajectory length.
    _, dX = x.shape

    d = x - t  # Error

    # Total cost
    # l = sum(log(cosh(x - t)) * w)
    l = np.sum(np.log(np.cosh(d)) * w, axis=1)

    # First order derivative
    # lx = tanh(x - t) * ws
    lx = np.tanh(d) * w

    # Second order derivative
    # lxx = w / cosh^2(x - t)
    lxx = np.einsum("ij,jk->ijk", w / np.cosh(d)**2, np.eye(dX))

    return l, lx, lxx



"""Batched versions of commin math operations."""
import numpy as np
from scipy.linalg import solve_triangular


def batched_inv_spd(a_chol: np.ndarray) -> np.ndarray:
    """Computes inverse of a batch of s.p.d. matrices from their cholesky decomposition.

    Exploits s.p.d.-ness for faster inverse.

    Args:
        a_chol: (T, dX, dX), batch of cholesky decompositions.

    """
    T, dX, _ = a_chol.shape
    a_inv = np.empty_like(a_chol)

    for t in range(T):
        a_inv[t] = solve_triangular(
            a_chol[t].T,
            solve_triangular(
                a_chol[t],
                np.eye(dX),
                lower=True,
                overwrite_b=True,
                check_finite=False,
            ),
            overwrite_b=True,
            check_finite=False,
        )
    return a_inv


def batched_cholesky(a: np.ndarray) -> np.ndarray:
    """Computes cholesky decomposition of a batch of s.p.d. matrices.

    Args:
        a: (T, dX, dX), batch of s.p.d. matrices

    Returns:
        Batch of lower triangular matrices.

    """
    T, _, _ = a.shape
    a_chol = np.empty_like(a)

    for t in range(T):
        a_chol[t] = np.linalg.cholesky(a[t])
    return a_chol


def symmetrize(A: np.ndarray) -> np.ndarray:
    """Symmetrizes a matrix or a batch of matrices to eliminate numerical errors.

    Modifies the given matrix in-place.
    """
    A += np.swapaxes(A, -1, -2)
    A /= 2
    return A


def regularize(A: np.ndarray, regularization) -> np.ndarray:
    """Regularizes a matrix or a batch of matrices by adding a constant to the diagonal.

    Modifies the given matrix in-place.
    """
    idx = np.arange(A.shape[-1])
    A[..., idx, idx] += regularization
    return A


def trace_of_product(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    """Compute the trace of a matrix product Tr(A @ B).

    This is an O(n²) operation, as it avoids to compute the actial matrix product.

    Args:
        A: (..., n, m) Matrix, or stack of matrices
        B: (..., m, n) Matrix, or stack of matrices

    Returns:
        trace: Trace of stack of traces
    """
    return np.einsum("...ij,...ji->...", A, B)


def batched_multivariate_normal(mean: np.ndarray, covar: np.ndarray, N: int, rng) -> np.ndarray:
    """Draw `N` samples from `T` multivariate normal distributions each.

    Args:
        mean: (T, dX)
        covar: (T, dX, dX)
        N: Number of samples
        rng: Random number generator

    Returns:
        X: (N, T, dX) Radnom samples
    """
    T, dX = mean.shape

    X = np.empty((N, T, dX))
    for t in range(T):
        X[:, t] = rng.multivariate_normal(mean[t], covar[t], N)

    return X



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator


def plot_line_set(x, ys, color=None, individual_lines="auto", label=None):
    N, _ = ys.shape

    # Only show individual lines if there are only a few in auto mode
    if individual_lines == "auto":
        individual_lines = N <= 5

    # Plot indiviadual data lines
    if individual_lines:
        for n in range(N):
            l = plt.plot(x, ys[n], c=color, linewidth=0.5)[0]
            color = l.get_color()  # Remember color

    # Plot distriution
    mean = np.mean(ys, axis=0)
    std = np.std(ys, axis=0)
    conf = 1.96 * std / np.sqrt(N)
    l = plt.plot(mean, c=color, label=label)[0]
    color = l.get_color()  # Remember color
    plt.fill_between(
        x,
        mean - conf,
        mean + conf,
        facecolor=color,
        alpha=0.25,
        interpolate=True,
    )


def visualize_trajectories(file, X, U):
    plt.figure(figsize=(16, 9))

    ax1 = plt.subplot(2, 1, 1)
    for d in range(X.shape[-1]):
        plot_line_set(np.arange(X.shape[1]), X[:, :, d], label=f"$x_{{{d}}}$")
    plt.ylabel("state")
    plt.legend()
    plt.grid(linestyle=":")
    ax1.xaxis.set_major_locator(MaxNLocator(integer=True))

    plt.subplot(2, 1, 2, sharex=ax1)
    for d in range(U.shape[-1]):
        plot_line_set(np.arange(U.shape[1]), U[:, :, d], label=f"$u_{{{d}}}$")
    plt.ylabel("action")
    plt.xlabel("$t$")
    plt.legend()
    plt.grid(linestyle=":")

    plt.tight_layout()
    plt.savefig(file)
    plt.close()



from donk.visualization.costs import visualize_costs
from donk.visualization.linear import (
    visualize_coefficients, visualize_covariance, visualize_linear_dynamics_model, visualize_linear_model, visualize_linear_policy,
    visualize_prediction, visualize_prediction_error, visualize_predictor_target_correlation, visualize_predictor_target_scatter
)
from donk.visualization.policy import visualize_policy_actions
from donk.visualization.states import visualize_correlation
from donk.visualization.traj_opt import visualize_iLQR
from donk.visualization.trajectories import visualize_trajectories

__all__ = [
    "visualize_linear_model",
    "visualize_coefficients",
    "visualize_covariance",
    "visualize_prediction",
    "visualize_prediction_error",
    "visualize_predictor_target_correlation",
    "visualize_predictor_target_scatter",
    "visualize_correlation",
    "visualize_iLQR",
    "visualize_trajectories",
    "visualize_linear_dynamics_model",
    "visualize_linear_policy",
    "visualize_policy_actions",
    "visualize_costs",
]



from __future__ import annotations

from abc import abstractmethod

import numpy as np

from donk.policy import Policy, smooth_noise
from donk.samples.sampler import Sampler


class GymSampler(Sampler):

    def __init__(self, env, smooth_kernel: float = 0.0) -> None:
        super().__init__()

        self.env = env
        self.smooth_kernel = smooth_kernel

    @abstractmethod
    def convert_observation(self, obs) -> np.ndarray:
        """Convert one observation from the Gym environment to a flat numpy array."""

    def convert_action(self, u: np.ndarray) -> np.ndarray:
        """Convert one action from a flat numpy array to one action for the from the Gym environment."""
        return u

    def take_sample(self, pol: Policy, T: int, condition: int, rng: np.random.Generator = None) -> tuple[np.ndarray, np.ndarray]:
        """Take one policy sample.

        Args:
            pol: Policy to sample
            T: Number of timesteps to sample
            condition: Initial condition to reset the environemnt into
            rng: Generator for action noise. `None` indicates no action noise.

        Returns:
            X: (T+1, dX) states
            U: (T, dU) states
        """
        dX, dU = pol.dX, pol.dU

        # Set initial condition
        self.env.seed(condition)

        # Allocate space
        X = np.empty((T + 1, dX))
        U = np.empty((T, dU))

        # Observe initial state
        obs = self.env.reset()
        X[0] = self.convert_observation(obs)

        # Generate action noise
        if rng is not None:
            noise = rng.standard_normal((T, dU))
            if self.smooth_kernel > 0:  # Apply smoothing
                noise = smooth_noise(noise, self.smooth_kernel)

        # Perform policy rollout
        for t in range(T):
            U[t] = pol.act(X[t], t, noise[t] if rng is not None else None)
            obs, _, done, _ = self.env.step(self.convert_action((U[t])))
            if done and t < T - 1:
                raise Exception(f'Iteration ended prematurely {t+1}/{T}')
            X[t + 1] = self.convert_observation(obs)

        return X, U



import numpy as np


class TrajectoryDistribution:
    """A normally distributed trajectory."""

    def __init__(self, mean: np.ndarray, covar: np.ndarray, dX: int, dU: int = None) -> None:
        """Initialize this `TrajectoryDistribution`.

        Args:
            mean: (..., T, dX+dU) Means of the trajectory distribution
            covar: (..., T, dX+dU, dX+dU) Covariances of the trajectory distribution
            dX: Dimension of state space
            dU: Dimension of action space, may be inferred if not explicitly stated
        """
        if dU is None:
            dU = mean.shape[-1] - dX

        # Check shapes
        assert mean.shape[-1:] == (dX + dU, ), f"{mean.shape[-1:]} != {(dX + dU,)}"
        assert covar.shape[-2:] == (dX + dU, dX + dU), f"{covar.shape[-2:]} != {(dX + dU, dX + dU)}"
        assert mean.shape[:-1] == covar.shape[:-2], f"{mean.shape[:-1]} != {covar.shape[:-2]}"

        self.mean = mean
        self.covar = covar
        self.dX = dX
        self.dU = dU

    @property
    def X_mean(self):
        """Get state component of trajectory means."""
        return self.mean[..., :self.dX]

    @property
    def X_covar(self):
        """Get state component of trajectory covariances."""
        return self.covar[..., :self.dX, :self.dX]

    @property
    def U_mean(self):
        """Get action component of trajectory means."""
        return self.mean[..., :-1, self.dX:]

    @property
    def U_covar(self):
        """Get action component of trajectory covariances."""
        return self.covar[..., :-1, self.dX:, self.dX:]



import numpy as np
from sklearn.mixture import GaussianMixture

from donk.dynamics.prior import DynamicsPrior, NormalInverseWishart


class GMMPrior(DynamicsPrior):
    """A Gaussian Mixture Model (GMM) based prior."""

    def __init__(self, n_clusters, random_state=None) -> None:
        self.gmm = GaussianMixture(n_components=n_clusters, random_state=random_state)

    def update(self, XUX: np.ndarray) -> None:
        """Update the prior.

        Args:
            XUX: (N, dX+dU+dX), transitions
        """
        self.N = XUX.shape[0]
        self.gmm.fit(XUX)

        # Enable warm start for further updates
        self.gmm.warm_start = True

    def eval(self, XUX: np.ndarray) -> NormalInverseWishart:
        """Evaluate the prior for the given transitions.

        Args:
            XUX: (N, dX+dU+dX), transitions
        """
        d = XUX.shape[1]
        wts = np.mean(self.gmm.predict_proba(XUX), axis=0)

        mu0 = np.sum(wts[:, np.newaxis] * self.gmm.means_, axis=0)
        Phi = np.sum(wts[:, np.newaxis, np.newaxis] * self.gmm.covariances_, axis=0)

        # Proportional number of samples in corresponding clusters
        N = np.sum(wts * self.gmm.weights_) * self.N

        return NormalInverseWishart.non_informative_prior(d).posterior(mu0, Phi, N)



from __future__ import annotations

from collections.abc import Callable
from typing import Dict, List

import numpy as np

from donk.costs.cost_function import CostFunction
from donk.costs.quadratic_costs import QuadraticCosts


def _vectorize_cost_function(fun):
    return np.vectorize(lambda X, U: np.array(fun(X, U)), signature="(v,x),(t,u)->(v)")


class SymbolicCostFunction(CostFunction):
    """CostFunction using symbolic differentiation using SymPy to approximate arbitrary functions."""

    def __init__(self, cost_fun: Callable[[np.ndarray, np.ndarray], np.ndarray], T: int, dX: int, dU: int) -> None:
        """Initialize this SymbolicCostFunction.

        Args:
            cost_fun: The primitive cost function. A callable which evaluates the costs for a trajectory at each timestep.
                      Returns an ndarray with shape (T, )
            T: Time horizon
            dX: Dimension of state space
            dU: Dimension of action space
        """
        from sympy import Matrix, diff, lambdify, symbols

        self.T, self.dX, self.dU = T, dX, dU

        # Create symbols
        X_sym = np.array(symbols(f"x:{(T+1)*dX}")).reshape(T + 1, dX)
        U_sym = np.array(symbols(f"u:{T*dU}")).reshape(T, dU)

        # Eval costs
        costs = cost_fun(X_sym, U_sym)

        # Store base cost function as vectorized to allow broadcasting
        self.cost_fun = _vectorize_cost_function(lambdify([X_sym, U_sym], costs))

        self.C = []
        self.c = []
        self.cc = []
        for t in range(T + 1):
            XU = np.concatenate([X_sym[t], U_sym[t]]) if t < T else X_sym[t]

            loss = costs[t]
            loss_d = np.array([diff(loss, xu) for xu in XU])
            loss_dd = np.array([[diff(loss_d[j], xu) for j in range(len(XU))] for xu in XU])

            # Lambdify sympy expressions for performance
            self.C.append(lambdify([X_sym, U_sym], Matrix(loss_dd)))
            self.c.append(lambdify([X_sym, U_sym], Matrix(loss_d - XU @ loss_dd)))
            self.cc.append(lambdify([X_sym, U_sym], loss - XU @ loss_d + XU.T @ loss_dd @ XU / 2))

    def quadratic_approximation(self, X: np.ndarray, U: np.ndarray) -> QuadraticCosts:
        """Compute a quadratic approximation (2nd order Taylor) at the given trajectory.

        Args:
            X: (T+1, dX), states
            U: (T, dX), actions
        """
        T, dX, dU = self.T, self.dX, self.dU

        # Check shapes
        assert X.shape == (T + 1, dX), f"{X.shape} != {(T + 1, dX)}"
        assert U.shape == (T, dU), f"{U.shape} != {(T , dU)}"

        # Allocate space
        C = np.empty((T + 1, dX + dU, dX + dU))
        c = np.empty((T + 1, dX + dU))
        cc = np.empty((T + 1, ))

        # Numerical evaluations for each timestep
        for t in range(T):
            C[t] = self.C[t](X, U)
            c[t] = self.c[t](X, U).flatten()
            cc[t] = self.cc[t](X, U)
        # Final state
        C[T] = 0  # Zero action potion
        C[T, :dX, :dX] = self.C[T](X, U)
        c[T, dX:] = 0  # Zero action potion
        c[T, :dX] = self.c[T](X, U).flatten()
        cc[T] = self.cc[T](X, U)

        return QuadraticCosts(C, c, cc)

    def compute_costs(self, X: np.ndarray, U: np.ndarray) -> np.ndarray:
        """Evaluate costs for trajectories.

        Args:
            X: (..., T+1, dX), states
            U: (..., T, dU), actions

        Returns:
            costs: (..., T+1), Costs at each time step
        """
        return self.cost_fun(X, U)


class MultipartSymbolicCostFunction(SymbolicCostFunction):
    """SymbolicCostFunction composed of named partial cost functions."""

    def __init__(
        self, cost_funs: List[Callable[[np.ndarray, np.ndarray], np.ndarray]], cost_function_names: List[str], T: int, dX: int, dU: int
    ) -> None:
        from sympy import lambdify, symbols

        def cost_fun(X, U):
            """Sum up individual parts."""
            cost = cost_funs[0](X, U)
            for fn in cost_funs[1:]:
                cost += fn(X, U)
            return cost

        super().__init__(cost_fun, T, dX, dU)

        # Create symbols
        X_sym = np.array(symbols(f"x:{(T+1)*dX}")).reshape(T + 1, dX)
        U_sym = np.array(symbols(f"u:{T*dU}")).reshape(T, dU)

        # Lambdify and vectorize cost functions
        self.cost_funs = [_vectorize_cost_function(lambdify([X_sym, U_sym], cf(X_sym, U_sym))) for cf in cost_funs]

        self.cost_function_names = cost_function_names

    def compute_costs_individual(self, X: np.ndarray, U: np.ndarray) -> Dict[str, np.ndarray]:
        """Evaluate costs for trajectories.

        Args:
            X: (..., T+1, dX), states
            U: (..., T, dU), actions

        Returns:
            costs: {name: (..., T+1)}, Costs at each time step, of each sub-cost function
        """
        return {name: cf(X, U) for cf, name in zip(self.cost_funs, self.cost_function_names)}



import numpy as np

from donk.dynamics import to_transitions


class TransitionPool:
    """Stores transition tuples."""

    def __init__(self, capacity: int = None) -> None:
        """Init this TransitionPool.

        Args:
            capacity: Site limit of this pool. Oldest transitions will be discarted first. `None` means no limit.
        """
        self._XUX = None
        self.capacity = capacity

    def add(self, X: np.ndarray, U: np.ndarray) -> None:
        """Add state-action sequences to the pool.

        Args:
            X: (N, T+1, dX), States
            U: (N, T, dU), Actions
        """
        # Check shapes
        assert X.shape[0] == U.shape[0], f"{X.shape[0]} != {U.shape[0]}"
        assert X.shape[1] == U.shape[1] + 1, f"{X.shape[1]} != {U.shape[1]+1}"
        if self._XUX is None:
            # Pool was empty, assume dimensions
            self.dX, self.dU = X.shape[-1], U.shape[-1]

            # Init pool
            self._XUX = to_transitions(X, U)
        else:
            # Sor safety, compare shapes before adding
            assert X.shape[-1] == self.dX, f"{X.shape[-1]} != {self.dX}"
            assert U.shape[-1] == self.dU, f"{U.shape[-1]} != {self.dU}"

            # Add to pool
            self._XUX = np.concatenate([self._XUX, to_transitions(X, U)], axis=0)
        if self.capacity is not None:
            self._XUX = self._XUX[-self.capacity:]

    def get_transitions(self, N: int = None):
        """Get transitions from the pool.

        Args:
            N: (optional) Return only the `N` newest transitions. Default is to return all.
        """
        XUX = self._XUX if N is None else self._XUX[-N:]

        # Create write-protected view
        XUX = XUX.view()
        XUX.flags.writeable = False

        return XUX



from pathlib import Path
from typing import List

import matplotlib.pyplot as plt
import numpy as np

from donk.policy import Policy


def visualize_policy_actions(
    output_file: Path,
    X: np.ndarray,
    policies: List[Policy],
    policy_labels: List[str],
    actions: List[int],
    action_labels: List[str],
    action_scaler=None,
):
    """Compares the actions of different policies of the same sates.

    Args:
        X: (N, T+1, dX) States
        policies: List of policies
        policy_labels: Names for the policies
        actions: Indices of actions to plot
        action_labels: Names for the actions
        action_scaler: sklearn Scaler
    """
    N, _, _ = X.shape
    T = X.shape[1] - 1

    # Compute actions
    U = []
    for pol in policies:
        u = pol.act(X[:, :-1], t=None)
        if action_scaler is not None:  # Inverse transform
            u = action_scaler.inverse_transform(u.reshape(N * T, pol.dU)).reshape(N, T, pol.dU)
        U.append(u)

    for j, action in enumerate(actions):
        plt.subplot(1, len(actions), j + 1)
        plt.title(action_labels[j])
        for i, u in enumerate(U):
            for n in range(N):
                plt.plot(u[n][:, action], color=f"C{i}", label=policy_labels[i] if n == 0 else None)

        plt.xlabel("$t$")
        plt.ylabel(r"$\mathbf{u}_t$")
        if j == len(actions) - 1:
            plt.legend()

    # Export
    plt.tight_layout()
    if output_file is not None:
        plt.savefig(output_file)
    else:
        plt.show()
    plt.close()



from __future__ import annotations

import numpy as np


class StateDistribution:
    """A normally distributed state."""

    def __init__(self, mean: np.ndarray, covar: np.ndarray) -> None:
        """Initialize this `StateDistribution`.

        Args:
            mean: (..., dX) Means of the state distribution
            covar: (..., dX, dX) Covariances of the state distribution
            dX: Dimension of state space
        """
        dX = mean.shape[-1]

        # Check shapes
        assert covar.shape[-2:] == (dX, dX), f"{covar.shape[-2:]} != {(dX, dX )}"
        assert mean.shape[:-1] == covar.shape[:-2], f"{mean.shape[:-1]} != {covar.shape[:-2]}"

        self.mean = mean
        self.covar = covar
        self.dX = dX

    @staticmethod
    def fit(X: np.ndarray) -> StateDistribution:
        """Fit a `StateDistribution` to some states.

        Args:
            X: (N, dX), states
        """
        return StateDistribution(
            mean=np.mean(X, axis=0),
            covar=np.cov(X, rowvar=False),
        )



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
from tqdm import tqdm

import donk.policy


class Sampler(ABC):

    @abstractmethod
    def take_sample(self, pol: donk.policy.Policy, T: int, condition, rng: np.random.Generator = None) -> tuple[np.ndarray, np.ndarray]:
        """Take one policy sample.

        Args:
            pol: Policy to sample
            T: Number of timesteps to sample
            condition: Initial condition to reset the environemnt into
            rng: Generator for action noise. `None` indicates no action noise.

        Returns:
            X: (T+1, dX) states
            U: (T, dU) states
        """

    def take_samples(self,
                     N: int,
                     pol: donk.policy.Policy,
                     T: int,
                     condition,
                     rng: np.random.Generator = None,
                     silent: bool = True) -> tuple[np.ndarray, np.ndarray]:
        """Takes multiple policy sample.

        Args:
            N: number of samples to take
            pol: Policy to sample
            T: Number of timesteps to sample
            condition: Initial condition to reset the environemnt into
            rng: Generator for action noise. `None` indicates no action noise.
            silent: Whether to show a tqdm progress bar

        Returns:
            X: (N, T+1, dX) states
            U: (N, T, dU) states
        """
        # Allocate space
        X = np.empty((N, T + 1, pol.dX))
        U = np.empty((N, T, pol.dU))

        # Take N samples
        for i in tqdm(range(N), desc="Sampling", disable=silent):
            X[i], U[i] = self.take_sample(pol, T, condition, rng)

        return X, U
